
static char rcsid[] = "@(#)$Id: mimesend.c,v 1.1.4.2 2011-04-07 04:44:41 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.1.4.2 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI>
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 ******************************************************************************
 * Moved needs_encoding() from src/mime.c. However needs_encoding() do not
 * exist on Elm 2.4. needs_encoding() may be introduced on ELM 2.4 PL24 MEx
 * series.
 *
 * That Elm 2.4 file src/mime.c was following copyright:
 *
 *  The Elm Mail System 
 *
 *			Copyright (c) 1988-1992 USENET Community Trust
 *			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/


#include "def_misc.h"

DEBUG_VAR(Debug,__FILE__,"mime");


/* Determine whether or not the data in "fp" needs to be QUOTED-PRINTABLE
    encoded.  Used when sending a message.
       1 == have 8bit data                (HAVE_8BIT)
       4 == have control charcters        (HAVE_CTRL)
       8 == mime requires encoding BINARY (HAVE_BINARY)
*/

int needs_encoding (fp,list)
     FILE *fp;
     struct scan_list *list;
{
    int ch;
    int ret = 0;
    int len = 0;
    int printed = 0;

    rewind (fp);

    while ((ch = fgetc (fp)) != EOF) {
	int need_more = 0;
	
	if (list) {
	    need_more = scanlist_need_more(list,ch);
	
	    if (!need_more && !printed) {
		DPRINT(Debug,9,(&Debug,    
				"\nneeds_encoding(): magic scan finished\n"));
		printed++;
	    }
	}

	/* check for end of line */
	if (ch == 13) {   /* CR */
	    ch = fgetc(fp);

	    if (ch != 10) {  /* Not CR LF */
		DPRINT(Debug,9,(&Debug,    
				"\nneeds_encoding(): found CR without LF\n"));
		ret |= HAVE_BINARY;
	    }
	    if (ch == EOF)
		break;

	    if (list) {
		need_more = scanlist_need_more(list,ch);
		
		if (!need_more && !printed) {
		    DPRINT(Debug,9,(&Debug,    
				    "\nneeds_encoding(): magic scan finished\n"));
		    printed++;
		}
	    }
	}
	if (ch == 10) {
	    len = 0;
	    continue; /* skip newlines and tabs */
	}
	len++;
	if (len > 990) {
	    DPRINT(Debug,9,(&Debug,   
			    "\nneeds_encoding(): Line over 990 characters\n"));	
	    ret |= HAVE_BINARY;
	}

	if (ch == 9) /* skip newlines and tabs */
	    continue;
	if (ch < 32 || ch > 126) {
	    DPRINT(Debug,9,(&Debug,    
			    "\nneeds_encoding(): found char decimal=%d\n", ch));
	    if (ch == 0) 
		ret |= HAVE_BINARY;
	    if (ch < 32 || ch == 127)
		ret |= HAVE_CTRL;
	    if (ch > 127)
		ret |= HAVE_8BIT;
	}

	if (ret == (HAVE_8BIT | HAVE_CTRL | HAVE_BINARY) &&
	    !need_more)
	    break;
	
    }
    rewind (fp);

    DPRINT(Debug,4,(&Debug,
		    "needs_encoding()=%d%s%s%s\n",
		    ret,
		    ret & HAVE_8BIT ? " HAVE_8BIT" : "",
		    ret & HAVE_CTRL ? " HAVE_CTRL" : "",
		    ret & HAVE_BINARY ? " HAVE_BINARY" : ""));
		  
    return (ret);
}


/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
